package com.sun.java.util.jar.pack;

import java.io.IOException;
import java.util.Arrays;
import java.util.Iterator;
import java.util.HashMap;
import java.io.ByteArrayOutputStream;
import java.util.Collection;
import java.util.Collections;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

class Attribute implements Comparable<Attribute>
{
    Attribute.Layout def;
    byte[] bytes;
    Object fixups;
    private static final Map<List<Attribute>, List<Attribute>> canonLists;
    private static final Map<Attribute.Layout, Attribute> attributes;
    private static final Map<Attribute.Layout, Attribute> standardDefs;
    static final byte EK_INT = 1;
    static final byte EK_BCI = 2;
    static final byte EK_BCO = 3;
    static final byte EK_FLAG = 4;
    static final byte EK_REPL = 5;
    static final byte EK_REF = 6;
    static final byte EK_UN = 7;
    static final byte EK_CASE = 8;
    static final byte EK_CALL = 9;
    static final byte EK_CBLE = 10;
    static final byte EF_SIGN = 1;
    static final byte EF_DELTA = 2;
    static final byte EF_NULL = 4;
    static final byte EF_BACK = 8;
    static final int NO_BAND_INDEX = -1;
    
    public String name() {
        return this.def.name();
    }
    
    public Attribute.Layout layout() {
        return this.def;
    }
    
    public byte[] bytes() {
        return this.bytes;
    }
    
    public int size() {
        return this.bytes.length;
    }
    
    public ConstantPool.Entry getNameRef() {
        return this.def.getNameRef();
    }
    
    private Attribute(Attribute attribute) {
        this.def = attribute.def;
        this.bytes = attribute.bytes;
        this.fixups = attribute.fixups;
    }
    
    public Attribute(Attribute.Layout def, byte[] bytes, Object fixups) {
        this.def = def;
        this.bytes = bytes;
        Fixups.setBytes(this.fixups = fixups, bytes);
    }
    
    public Attribute(Attribute.Layout layout, byte[] array) {
        this(layout, array, null);
    }
    
    public Attribute addContent(byte[] bytes, Object fixups) {
        assert !(!this.isCanonical());
        if (bytes.length == 0 && fixups == null) {
            return this;
        }
        Attribute attribute = new Attribute(this);
        attribute.bytes = bytes;
        Fixups.setBytes(attribute.fixups = fixups, bytes);
        return attribute;
    }
    
    public Attribute addContent(byte[] array) {
        return this.addContent(array, null);
    }
    
    public void finishRefs(ConstantPool.Index index) {
        if (this.fixups != null) {
            Fixups.finishRefs(this.fixups, this.bytes, index);
            this.fixups = null;
        }
    }
    
    public boolean isCanonical() {
        return this == this.def.canon;
    }
    
    @Override
    public int compareTo(Attribute attribute) {
        return this.def.compareTo(attribute.def);
    }
    
    public static List<Attribute> getCanonList(List<Attribute> list) {
        synchronized (Attribute.canonLists) {
            List<Attribute> unmodifiableList = (List<Attribute>)Attribute.canonLists.get(list);
            if (unmodifiableList == null) {
                ArrayList<Object> list2 = new ArrayList(list.size());
                list2.addAll(list);
                unmodifiableList = Collections.unmodifiableList(list2);
                Attribute.canonLists.put(list, unmodifiableList);
            }
            return unmodifiableList;
        }
    }
    
    public static Attribute find(int n, String s, String s2) {
        Attribute.Layout key = Attribute.Layout.makeKey(n, s, s2);
        synchronized (Attribute.attributes) {
            Attribute canonicalInstance = (Attribute)Attribute.attributes.get(key);
            if (canonicalInstance == null) {
                canonicalInstance = new Attribute.Layout(n, s, s2).canonicalInstance();
                Attribute.attributes.put(key, canonicalInstance);
            }
            return canonicalInstance;
        }
    }
    
    public static Attribute.Layout keyForLookup(int n, String s) {
        return Attribute.Layout.makeKey(n, s);
    }
    
    public static Attribute lookup(Map<Attribute.Layout, Attribute> standardDefs, int n, String s) {
        if (standardDefs == null) {
            standardDefs = Attribute.standardDefs;
        }
        return (Attribute)standardDefs.get(Attribute.Layout.makeKey(n, s));
    }
    
    public static Attribute define(Map<Attribute.Layout, Attribute> map, int n, String s, String s2) {
        Attribute find = find(n, s, s2);
        map.put(Attribute.Layout.makeKey(n, s), find);
        return find;
    }
    
    public static String contextName(int n) {
        switch (n) {
            case 0: {
                return "class";
            }
            case 1: {
                return "field";
            }
            case 2: {
                return "method";
            }
            case 3: {
                return "code";
            }
            default: {
                return null;
            }
        }
    }
    
    void visitRefs(Attribute.Holder holder, int n, final Collection<ConstantPool.Entry> collection) {
        if (n == 0) {
            collection.add(this.getNameRef());
        }
        if (this.bytes.length == 0) {
            return;
        }
        if (!this.def.hasRefs) {
            return;
        }
        if (this.fixups != null) {
            Fixups.visitRefs(this.fixups, collection);
            return;
        }
        this.def.parse(holder, this.bytes, 0, this.bytes.length, new Attribute.ValueStream() {
            @Override
            public void putInt(int n, int n2) {
            }
            
            @Override
            public void putRef(int n, ConstantPool.Entry entry) {
                collection.add(entry);
            }
            
            @Override
            public int encodeBCI(int n) {
                return n;
            }
        });
        return;
    }
    
    public void parse(Attribute.Holder holder, byte[] array, int n, int n2, Attribute.ValueStream valueStream) {
        this.def.parse(holder, array, n, n2, valueStream);
    }
    
    public Object unparse(Attribute.ValueStream valueStream, ByteArrayOutputStream byteArrayOutputStream) {
        return this.def.unparse(valueStream, byteArrayOutputStream);
    }
    
    @Override
    public String toString() {
        return this.def + "{" + ((this.bytes == null) ? -1 : this.size()) + "}" + ((this.fixups == null) ? "" : this.fixups.toString());
    }
    
    public static String normalizeLayoutString(String s) {
        StringBuilder sb = new StringBuilder();
        int i = 0;
        int length = s.length();
        while (i < length) {
            int char1 = s.charAt(i++);
            if (char1 <= ' ') {
                continue;
            }
            if (char1 == '#') {
                int index = s.indexOf(10, i);
                int index2 = s.indexOf(13, i);
                if (index < 0) {
                    index = length;
                }
                if (index2 < 0) {
                    index2 = length;
                }
                i = Math.min(index, index2);
            }
            else if (char1 == '\\') {
                sb.append((int)s.charAt(i++));
            }
            else if (char1 == '0' && s.startsWith("0x", i - 1)) {
                int beginIndex = i - 1;
                int j = beginIndex + 2;
                while (j < length) {
                    int char2 = s.charAt(j);
                    if ((char2 >= '0' && char2 <= '9') || (char2 >= 'a' && char2 <= 'f')) {
                        ++j;
                    }
                    else {
                        break;
                    }
                }
                if (j > beginIndex) {
                    sb.append(Integer.decode(s.substring(beginIndex, j)));
                    i = j;
                }
                else {
                    sb.append((char)char1);
                }
            }
            else {
                sb.append((char)char1);
            }
        }
        return sb.toString();
    }
    
    static Attribute.Layout.Element[] tokenizeLayout(Attribute.Layout layout, int n, String s) {
        ArrayList list = new ArrayList(s.length());
        tokenizeLayout(layout, n, s, list);
        Attribute.Layout.Element[] array = new Attribute.Layout.Element[list.size()];
        list.toArray(array);
        return array;
    }
    
    static void tokenizeLayout(Attribute.Layout layout, int n, String s, List<Attribute.Layout.Element> list) {
        int n2 = 0;
        int length = s.length();
        int i = 0;
        while (i < length) {
            Attribute.Layout.Element o = layout.new Element();
            int n3;
            switch (s.charAt(i++)) {
                case 'B':
                case 'H':
                case 'I':
                case 'V': {
                    n3 = 1;
                    --i;
                    i = tokenizeUInt(o, s, i);
                    break;
                }
                case 'S': {
                    n3 = 1;
                    --i;
                    i = tokenizeSInt(o, s, i);
                    break;
                }
                case 'P': {
                    n3 = 2;
                    if (s.charAt(i++) == 'O') {
                        o.flags |= 0x2;
                        if (n2 == 0) {
                            i = -i;
                            continue;
                        }
                        ++i;
                    }
                    --i;
                    i = tokenizeUInt(o, s, i);
                    break;
                }
                case 'O': {
                    n3 = 3;
                    o.flags |= 0x2;
                    if (n2 == 0) {
                        i = -i;
                        continue;
                    }
                    i = tokenizeSInt(o, s, i);
                    break;
                }
                case 'F': {
                    n3 = 4;
                    i = tokenizeUInt(o, s, i);
                    break;
                }
                case 'N': {
                    n3 = 5;
                    int tokenizeUInt = tokenizeUInt(o, s, i);
                    if (s.charAt(tokenizeUInt++) != '[') {
                        i = -tokenizeUInt;
                        continue;
                    }
                    int beginIndex;
                    i = skipBody(s, beginIndex = tokenizeUInt);
                    o.body = tokenizeLayout(layout, n, s.substring(beginIndex, i++));
                    break;
                }
                case 'T': {
                    n3 = 7;
                    i = tokenizeSInt(o, s, i);
                    ArrayList o2 = new ArrayList();
                    while (s.charAt(i++) == '(') {
                        i = s.indexOf(41, i);
                        String substring = s.substring(i, i++);
                        int length2 = substring.length();
                        if (s.charAt(i++) != '[') {
                            i = -i;
                        }
                        else {
                            if (s.charAt(i) == ']') {
                            }
                            else {
                                i = skipBody(s, i = i);
                            }
                            Attribute.Layout.Element[] tokenizeLayout = tokenizeLayout(layout, n, s.substring(i, i++));
                            if (length2 == 0) {
                                Attribute.Layout.Element o4 = layout.new Element();
                                o4.body = tokenizeLayout;
                                o4.kind = 8;
                                o4.removeBand();
                                o2.add(o4);
                            }
                            else {
                                int n4 = 1;
                                int n5 = 0;
                                while (true) {
                                    int index = substring.indexOf(44, n5);
                                    if (index < 0) {
                                        index = length2;
                                    }
                                    String substring2 = substring.substring(n5, index);
                                    if (substring2.length() == 0) {
                                        substring2 = "empty";
                                    }
                                    int caseDash = findCaseDash(substring2, 0);
                                    int value;
                                    int intAfter;
                                    if (caseDash >= 0) {
                                        value = parseIntBefore(substring2, caseDash);
                                        intAfter = parseIntAfter(substring2, caseDash);
                                        if (value >= intAfter) {
                                            i = -i;
                                            break;
                                        }
                                    }
                                    else {
                                        intAfter = (value = Integer.parseInt(substring2));
                                    }
                                    while (true) {
                                        Attribute.Layout.Element o3 = layout.new Element();
                                        o3.body = tokenizeLayout;
                                        o3.kind = 8;
                                        o3.removeBand();
                                        if (n4 == 0) {
                                            o3.flags |= 0x8;
                                        }
                                        n4 = 0;
                                        o3.value = value;
                                        o2.add(o3);
                                        if (value == intAfter) {
                                            break;
                                        }
                                        ++value;
                                    }
                                    if (index == length2) {
                                        break;
                                    }
                                    n5 = index + 1;
                                }
                                continue;
                            }
                        }
                        Label_0889: {
                            o2.toArray(o.body = new Attribute.Layout.Element[o2.size()]);
                        }
                        o.kind = n3;
                        int j = 0;
                        while (j < o.body.length - 1) {
                            Attribute.Layout.Element element = o.body[j];
                            if (matchCase(o, element.value) != element) {
                                i = -i;
                                break;
                            }
                            ++j;
                        }
                        break;
                    }
                    i = -i;
                    break;
                }
                case '(': {
                    n3 = 9;
                    o.removeBand();
                    i = s.indexOf(41, i);
                    String substring3 = s.substring(i + 1, i++);
                    int int1 = Integer.parseInt(substring3);
                    int value2 = n + int1;
                    if (!(int1 + "").equals(substring3) || layout.elems == null || value2 < 0 || value2 >= layout.elems.length) {
                        i = -i;
                        continue;
                    }
                    Attribute.Layout.Element element2 = layout.elems[value2];
                    assert element2.kind == 10;
                    o.value = value2;
                    o.body = new Attribute.Layout.Element[] { element2 };
                    if (int1 <= 0) {
                        o.flags |= 0x8;
                        element2.flags |= 0x8;
                        break;
                    }
                    break;
                }
                case 'K': {
                    n3 = 6;
                    switch (s.charAt(i++)) {
                        case 'I': {
                            o.refKind = 3;
                            break;
                        }
                        case 'J': {
                            o.refKind = 5;
                            break;
                        }
                        case 'F': {
                            o.refKind = 4;
                            break;
                        }
                        case 'D': {
                            o.refKind = 6;
                            break;
                        }
                        case 'S': {
                            o.refKind = 8;
                            break;
                        }
                        case 'Q': {
                            o.refKind = 53;
                            break;
                        }
                        case 'M': {
                            o.refKind = 15;
                            break;
                        }
                        case 'T': {
                            o.refKind = 16;
                            break;
                        }
                        case 'L': {
                            o.refKind = 51;
                            break;
                        }
                        default: {
                            i = -i;
                            continue;
                        }
                    }
                    break;
                }
                case 'R': {
                    n3 = 6;
                    switch (s.charAt(i++)) {
                        case 'C': {
                            o.refKind = 7;
                            break;
                        }
                        case 'S': {
                            o.refKind = 13;
                            break;
                        }
                        case 'D': {
                            o.refKind = 12;
                            break;
                        }
                        case 'F': {
                            o.refKind = 9;
                            break;
                        }
                        case 'M': {
                            o.refKind = 10;
                            break;
                        }
                        case 'I': {
                            o.refKind = 11;
                            break;
                        }
                        case 'U': {
                            o.refKind = 1;
                            break;
                        }
                        case 'Q': {
                            o.refKind = 50;
                            break;
                        }
                        case 'Y': {
                            o.refKind = 18;
                            break;
                        }
                        case 'B': {
                            o.refKind = 17;
                            break;
                        }
                        case 'N': {
                            o.refKind = 52;
                            break;
                        }
                        default: {
                            i = -i;
                            continue;
                        }
                    }
                    break;
                }
                default: {
                    i = -i;
                    continue;
                }
            }
            if (n3 == 6) {
                if (s.charAt(i++) == 'N') {
                    o.flags |= 0x4;
                    ++i;
                }
                --i;
                i = tokenizeUInt(o, s, i);
                layout.hasRefs = true;
            }
            n2 = (n3 == 2);
            o.kind = n3;
            o.layout = s.substring(i, i);
            list.add(o);
        }
    }
    
    static String[] splitBodies(String s) {
        ArrayList<String> list = new ArrayList();
        for (int i = 0; i < s.length(); ++i) {
            if (s.charAt(i++) != '[') {
                s.charAt(-i);
            }
            int beginIndex;
            i = skipBody(s, beginIndex = i);
            list.add(s.substring(beginIndex, i));
        }
        String[] array = new String[list.size()];
        list.toArray(array);
        return array;
    }
    
    private static int skipBody(String s, int n) {
        assert s.charAt(n - 1) == '[';
        if (s.charAt(n) == ']') {
            return -n;
        }
        int i = 1;
        while (i > 0) {
            switch (s.charAt(n++)) {
                case '[': {
                    ++i;
                    continue;
                }
                case ']': {
                    --i;
                    continue;
                }
            }
        }
        --n;
        assert s.charAt(n) == ']';
        return n;
    }
    
    private static int tokenizeUInt(Attribute.Layout.Element element, String s, int n) {
        switch (s.charAt(n++)) {
            case 'V': {
                element.len = 0;
                break;
            }
            case 'B': {
                element.len = 1;
                break;
            }
            case 'H': {
                element.len = 2;
                break;
            }
            case 'I': {
                element.len = 4;
                break;
            }
            default: {
                return -n;
            }
        }
        return n;
    }
    
    private static int tokenizeSInt(Attribute.Layout.Element element, String s, int index) {
        if (s.charAt(index) == 'S') {
            element.flags |= 0x1;
            ++index;
        }
        return tokenizeUInt(element, s, index);
    }
    
    private static boolean isDigit(char c) {
        return c >= '0' && c <= '9';
    }
    
    static int findCaseDash(String s, int fromIndex) {
        if (fromIndex <= 0) {
            fromIndex = 1;
        }
        int n = s.length() - 2;
        while (true) {
            int index = s.indexOf(45, fromIndex);
            if (index < 0 || index > n) {
                return -1;
            }
            if (isDigit(s.charAt(index - 1))) {
                int n2 = s.charAt(index + 1);
                if (n2 == '-' && index + 2 < s.length()) {
                    n2 = s.charAt(index + 2);
                }
                if (isDigit((char)n2)) {
                    return index;
                }
            }
            fromIndex = index + 1;
        }
    }
    
    static int parseIntBefore(String s, int n) {
        for (n = (n = n); n > 0 && isDigit(s.charAt(n - 1)); --n) {}
        if (n == n) {
            return Integer.parseInt("empty");
        }
        if (n >= 1 && s.charAt(n - 1) == '-') {
            --n;
        }
        assert !isDigit(s.charAt(n - 1));
        return Integer.parseInt(s.substring(n, n));
    }
    
    static int parseIntAfter(String s, int n) {
        int n2 = n + 1;
        int length = s.length();
        if (n2 < length && s.charAt(n2) == '-') {
            ++n2;
        }
        while (n2 < length && isDigit(s.charAt(n2))) {
            ++n2;
        }
        if (n2 == n2) {
            return Integer.parseInt("empty");
        }
        return Integer.parseInt(s.substring(n2, n2));
    }
    
    static String expandCaseDashNotation(String s) {
        int endIndex = findCaseDash(s, 0);
        if (endIndex < 0) {
            return s;
        }
        StringBuilder sb = new StringBuilder(s.length() * 3);
        int n = 0;
        while (true) {
            sb.append(s.substring(n, endIndex));
            n = endIndex + 1;
            int intBefore = parseIntBefore(s, endIndex);
            int intAfter = parseIntAfter(s, endIndex);
            assert intBefore < intAfter;
            sb.append(",");
            for (int i = intBefore + 1; i < intAfter; ++i) {
                sb.append(i);
                sb.append(",");
            }
            endIndex = findCaseDash(s, n);
            if (endIndex < 0) {
                sb.append(s.substring(n));
                return sb.toString();
            }
            continue;
        }
    }
    
    static int parseUsing(Attribute.Layout.Element[] array, Attribute.Holder holder, byte[] array2, int n, int n2, Attribute.ValueStream valueStream) {
        int n3 = 0;
        int n4 = 0;
        int n5 = n + n2;
        int[] array3 = { 0 };
        for (int i = 0; i < array.length; ++i) {
            Attribute.Layout.Element element = array[i];
            int bandIndex = element.bandIndex;
            switch (element.kind) {
                case 1: {
                    n = parseInt(element, array2, n, array3);
                    valueStream.putInt(bandIndex, array3[0]);
                    break;
                }
                case 2: {
                    n = parseInt(element, array2, n, array3);
                    int n6 = array3[0];
                    int encodeBCI = valueStream.encodeBCI(n6);
                    int n7;
                    if (!element.flagTest((byte)2)) {
                        n7 = encodeBCI;
                    }
                    else {
                        n7 = encodeBCI - n4;
                    }
                    n3 = n6;
                    n4 = encodeBCI;
                    valueStream.putInt(bandIndex, n7);
                    break;
                }
                case 3: {
                    assert !(!element.flagTest((byte)2));
                    n = parseInt(element, array2, n, array3);
                    int n8 = n3 + array3[0];
                    int encodeBCI2 = valueStream.encodeBCI(n8);
                    int n9 = encodeBCI2 - n4;
                    n3 = n8;
                    n4 = encodeBCI2;
                    valueStream.putInt(bandIndex, n9);
                    break;
                }
                case 4: {
                    n = parseInt(element, array2, n, array3);
                    valueStream.putInt(bandIndex, array3[0]);
                    break;
                }
                case 5: {
                    n = parseInt(element, array2, n, array3);
                    int n10 = array3[0];
                    valueStream.putInt(bandIndex, n10);
                    for (int j = 0; j < n10; ++j) {
                        n = parseUsing(element.body, holder, array2, n, n5 - n, valueStream);
                    }
                    break;
                }
                case 7: {
                    n = parseInt(element, array2, n, array3);
                    int n11 = array3[0];
                    valueStream.putInt(bandIndex, n11);
                    n = parseUsing(matchCase(element, n11).body, holder, array2, n, n5 - n, valueStream);
                    break;
                }
                case 9: {
                    assert element.body.length == 1;
                    assert element.body[0].kind == 10;
                    if (element.flagTest((byte)8)) {
                        valueStream.noteBackCall(element.value);
                    }
                    n = parseUsing(element.body[0].body, holder, array2, n, n5 - n, valueStream);
                    break;
                }
                case 6: {
                    n = parseInt(element, array2, n, array3);
                    int n12 = array3[0];
                    ConstantPool.Entry signatureEntry;
                    if (n12 == 0) {
                        signatureEntry = null;
                    }
                    else {
                        ConstantPool.Entry[] cpMap = holder.getCPMap();
                        signatureEntry = ((n12 >= 0 && n12 < cpMap.length) ? cpMap[n12] : null);
                        int refKind = element.refKind;
                        if (signatureEntry != null && refKind == 13 && signatureEntry.getTag() == 1) {
                            signatureEntry = ConstantPool.getSignatureEntry(signatureEntry.stringValue());
                        }
                        String str = (signatureEntry == null) ? "invalid CP index" : ("type=" + ConstantPool.tagName((int)signatureEntry.tag));
                        if (signatureEntry == null || !signatureEntry.tagMatches(refKind)) {
                            throw new IllegalArgumentException("Bad constant, expected type=" + ConstantPool.tagName(refKind) + " got " + str);
                        }
                    }
                    valueStream.putRef(bandIndex, signatureEntry);
                    break;
                }
                default: {
                    assert false;
                    break;
                }
            }
        }
        return n;
    }
    
    static Attribute.Layout.Element matchCase(Attribute.Layout.Element element, int n) {
        assert element.kind == 7;
        int n2 = element.body.length - 1;
        int i = 0;
        while (i < n2) {
            Attribute.Layout.Element element2 = element.body[i];
            assert element2.kind == 8;
            if (n == element2.value) {
                return element2;
            }
            ++i;
        }
        return element.body[n2];
    }
    
    private static int parseInt(Attribute.Layout.Element element, byte[] array, int n, int[] array2) {
        int n2 = 0;
        int n3 = element.len * 8;
        while (true) {
            n3 -= 8;
            if (n3 >= 0) {
                n2 += (array[n++] & 0xFF) << n3;
            }
            else {
                break;
            }
        }
        if (n3 < 32 && element.flagTest((byte)1)) {
            int n4 = 32 - n3;
            n2 = n2 << n4 >> n4;
        }
        array2[0] = n2;
        return n;
    }
    
    static void unparseUsing(Attribute.Layout.Element[] array, Object[] array2, Attribute.ValueStream valueStream, ByteArrayOutputStream byteArrayOutputStream) {
        int n = 0;
        int n2 = 0;
        for (int i = 0; i < array.length; ++i) {
            Attribute.Layout.Element element = array[i];
            int bandIndex = element.bandIndex;
            switch (element.kind) {
                case 1: {
                    unparseInt(element, valueStream.getInt(bandIndex), byteArrayOutputStream);
                    break;
                }
                case 2: {
                    int int1 = valueStream.getInt(bandIndex);
                    if (!element.flagTest((byte)2)) {
                    }
                    else {
                        int1 += n2;
                    }
                    assert n == valueStream.decodeBCI(n2);
                    int decodeBCI = valueStream.decodeBCI(int1);
                    unparseInt(element, decodeBCI, byteArrayOutputStream);
                    n = decodeBCI;
                    n2 = int1;
                    break;
                }
                case 3: {
                    int int2 = valueStream.getInt(bandIndex);
                    assert !(!element.flagTest((byte)2));
                    assert n == valueStream.decodeBCI(n2);
                    int n3 = n2 + int2;
                    int decodeBCI2 = valueStream.decodeBCI(n3);
                    unparseInt(element, decodeBCI2 - n, byteArrayOutputStream);
                    n = decodeBCI2;
                    n2 = n3;
                    break;
                }
                case 4: {
                    unparseInt(element, valueStream.getInt(bandIndex), byteArrayOutputStream);
                    break;
                }
                case 5: {
                    int int3 = valueStream.getInt(bandIndex);
                    unparseInt(element, int3, byteArrayOutputStream);
                    for (int j = 0; j < int3; ++j) {
                        unparseUsing(element.body, array2, valueStream, byteArrayOutputStream);
                    }
                    break;
                }
                case 7: {
                    int int4 = valueStream.getInt(bandIndex);
                    unparseInt(element, int4, byteArrayOutputStream);
                    unparseUsing(matchCase(element, int4).body, array2, valueStream, byteArrayOutputStream);
                    break;
                }
                case 9: {
                    assert element.body.length == 1;
                    assert element.body[0].kind == 10;
                    unparseUsing(element.body[0].body, array2, valueStream, byteArrayOutputStream);
                    break;
                }
                case 6: {
                    ConstantPool.Entry ref = valueStream.getRef(bandIndex);
                    int n4;
                    if (ref != null) {
                        array2[0] = Fixups.addRefWithLoc(array2[0], byteArrayOutputStream.size(), ref);
                        n4 = 0;
                    }
                    else {
                        n4 = 0;
                    }
                    unparseInt(element, n4, byteArrayOutputStream);
                    break;
                }
                default: {
                    assert false;
                    break;
                }
            }
        }
    }
    
    private static void unparseInt(Attribute.Layout.Element element, int i, ByteArrayOutputStream byteArrayOutputStream) {
        int n = element.len * 8;
        if (n == 0) {
            return;
        }
        if (n < 32) {
            int n2 = 32 - n;
            int n3;
            if (element.flagTest((byte)1)) {
                n3 = i << n2 >> n2;
            }
            else {
                n3 = i << n2 >>> n2;
            }
            if (n3 != i) {
                throw new InternalError("cannot code in " + (int)element.len + " bytes: " + i);
            }
        }
        while (true) {
            n -= 8;
            if (n >= 0) {
                byteArrayOutputStream.write((int)(byte)(i >>> n));
            }
            else {
                break;
            }
        }
        return;
    }
    
    static {
        canonLists = new HashMap();
        attributes = new HashMap();
        standardDefs = new HashMap();
        Map<Attribute.Layout, Attribute> standardDefs2 = Attribute.standardDefs;
        define(standardDefs2, 0, "Signature", "RSH");
        define(standardDefs2, 0, "Synthetic", "");
        define(standardDefs2, 0, "Deprecated", "");
        define(standardDefs2, 0, "SourceFile", "RUH");
        define(standardDefs2, 0, "EnclosingMethod", "RCHRDNH");
        define(standardDefs2, 0, "InnerClasses", "NH[RCHRCNHRUNHFH]");
        define(standardDefs2, 0, "BootstrapMethods", "NH[RMHNH[KLH]]");
        define(standardDefs2, 1, "Signature", "RSH");
        define(standardDefs2, 1, "Synthetic", "");
        define(standardDefs2, 1, "Deprecated", "");
        define(standardDefs2, 1, "ConstantValue", "KQH");
        define(standardDefs2, 2, "Signature", "RSH");
        define(standardDefs2, 2, "Synthetic", "");
        define(standardDefs2, 2, "Deprecated", "");
        define(standardDefs2, 2, "Exceptions", "NH[RCH]");
        define(standardDefs2, 2, "MethodParameters", "NB[RUNHFH]");
        define(standardDefs2, 3, "StackMapTable", "[NH[(1)]][TB(64-127)[(2)](247)[(1)(2)](248-251)[(1)](252)[(1)(2)](253)[(1)(2)(2)](254)[(1)(2)(2)(2)](255)[(1)NH[(2)]NH[(2)]]()[]][H][TB(7)[RCH](8)[PH]()[]]");
        define(standardDefs2, 3, "LineNumberTable", "NH[PHH]");
        define(standardDefs2, 3, "LocalVariableTable", "NH[PHOHRUHRSHH]");
        define(standardDefs2, 3, "LocalVariableTypeTable", "NH[PHOHRUHRSHH]");
        String[] array = { normalizeLayoutString("\n  # parameter_annotations :=\n  [ NB[(1)] ]     # forward call to annotations"), normalizeLayoutString("\n  # annotations :=\n  [ NH[(1)] ]     # forward call to annotation\n  "), normalizeLayoutString("\n  # annotation :=\n  [RSH\n    NH[RUH (1)]   # forward call to value\n    ]"), normalizeLayoutString("\n  # value :=\n  [TB # Callable 2 encodes one tagged value.\n    (\\B,\\C,\\I,\\S,\\Z)[KIH]\n    (\\D)[KDH]\n    (\\F)[KFH]\n    (\\J)[KJH]\n    (\\c)[RSH]\n    (\\e)[RSH RUH]\n    (\\s)[RUH]\n    (\\[)[NH[(0)]] # backward self-call to value\n    (\\@)[RSH NH[RUH (0)]] # backward self-call to value\n    ()[] ]") };
        String[] array2 = { normalizeLayoutString("\n # type-annotations :=\n  [ NH[(1)(2)(3)] ]     # forward call to type-annotations"), normalizeLayoutString("\n  # type-annotation :=\n  [TB\n    (0-1) [B] # {CLASS, METHOD}_TYPE_PARAMETER\n    (16) [FH] # CLASS_EXTENDS\n    (17-18) [BB] # {CLASS, METHOD}_TYPE_PARAMETER_BOUND\n    (19-21) [] # FIELD, METHOD_RETURN, METHOD_RECEIVER\n    (22) [B] # METHOD_FORMAL_PARAMETER\n    (23) [H] # THROWS\n    (64-65) [NH[PHOHH]] # LOCAL_VARIABLE, RESOURCE_VARIABLE\n    (66) [H] # EXCEPTION_PARAMETER\n    (67-70) [PH] # INSTANCEOF, NEW, {CONSTRUCTOR, METHOD}_REFERENCE_RECEIVER\n    (71-75) [PHB] # CAST, {CONSTRUCTOR,METHOD}_INVOCATION_TYPE_ARGUMENT, {CONSTRUCTOR, METHOD}_REFERENCE_TYPE_ARGUMENT\n    ()[] ]"), normalizeLayoutString("\n # type-path\n [ NB[BB] ]") };
        Map<Attribute.Layout, Attribute> standardDefs3 = Attribute.standardDefs;
        String s = array[3];
        String string = array[1] + array[2] + array[3];
        String string2 = array[0] + string;
        String string3 = array2[0] + array2[1] + array2[2] + array[2] + array[3];
        for (int i = 0; i < 4; ++i) {
            if (i != 3) {
                define(standardDefs3, i, "RuntimeVisibleAnnotations", string);
                define(standardDefs3, i, "RuntimeInvisibleAnnotations", string);
                if (i == 2) {
                    define(standardDefs3, i, "RuntimeVisibleParameterAnnotations", string2);
                    define(standardDefs3, i, "RuntimeInvisibleParameterAnnotations", string2);
                    define(standardDefs3, i, "AnnotationDefault", s);
                }
            }
            define(standardDefs3, i, "RuntimeVisibleTypeAnnotations", string3);
            define(standardDefs3, i, "RuntimeInvisibleTypeAnnotations", string3);
        }
        assert !(!expandCaseDashNotation("1-5").equals("1,2,3,4,5"));
        assert !(!expandCaseDashNotation("-2--1").equals("-2,-1"));
        assert !(!expandCaseDashNotation("-2-1").equals("-2,-1,0,1"));
        assert !(!expandCaseDashNotation("-1-0").equals("-1,0"));
        return;
    }
    
    public abstract static class Holder
    {
        protected int flags;
        protected List<Attribute> attributes;
        static final List<Attribute> noAttributes;
        
        protected abstract ConstantPool.Entry[] getCPMap();
        
        public int attributeSize() {
            return (this.attributes == null) ? 0 : this.attributes.size();
        }
        
        public void trimToSize() {
            if (this.attributes == null) {
                return;
            }
            if (this.attributes.isEmpty()) {
                this.attributes = null;
                return;
            }
            if ((this.attributes instanceof ArrayList)) {
                ArrayList<Attribute> list = (ArrayList)this.attributes;
                list.trimToSize();
                boolean n = true;
                for (Attribute attribute : list) {
                    if (!attribute.isCanonical()) {
                        n = false;
                    }
                    if (attribute.fixups != null) {
                        assert !attribute.isCanonical();
                        attribute.fixups = Fixups.trimToSize(attribute.fixups);
                    }
                    else {
                        continue;
                    }
                }
                if (!(n == 0)) {
                    this.attributes = Attribute.getCanonList(list);
                }
            }
            return;
        }
        
        public void addAttribute(Attribute attribute) {
            if (this.attributes == null) {
                this.attributes = new ArrayList(3);
            }
            else if (!(this.attributes instanceof ArrayList)) {
                this.attributes = new ArrayList(this.attributes);
            }
            this.attributes.add(attribute);
        }
        
        public Attribute removeAttribute(Attribute attribute) {
            if (this.attributes == null) {
                return null;
            }
            if (!this.attributes.contains(attribute)) {
                return null;
            }
            if (!(this.attributes instanceof ArrayList)) {
                this.attributes = new ArrayList(this.attributes);
            }
            this.attributes.remove(attribute);
            return attribute;
        }
        
        public Attribute getAttribute(int n) {
            return (Attribute)this.attributes.get(n);
        }
        
        protected void visitRefs(int n, Collection<ConstantPool.Entry> collection) {
            if (this.attributes == null) {
                return;
            }
            Iterator<Attribute> iterator = this.attributes.iterator();
            while (iterator.hasNext()) {
                ((Attribute)iterator.next()).visitRefs(this, n, collection);
            }
            return;
        }
        
        public List<Attribute> getAttributes() {
            if (this.attributes == null) {
                return Attribute.Holder.noAttributes;
            }
            return this.attributes;
        }
        
        public void setAttributes(List<Attribute> attributes) {
            if (attributes.isEmpty()) {
                this.attributes = null;
            }
            else {
                this.attributes = attributes;
            }
        }
        
        public Attribute getAttribute(String anObject) {
            if (this.attributes == null) {
                return null;
            }
            for (Attribute attribute : this.attributes) {
                if (attribute.name().equals(anObject)) {
                    return attribute;
                }
                continue;
            }
            return null;
        }
        
        public Attribute getAttribute(Attribute.Layout layout) {
            if (this.attributes == null) {
                return null;
            }
            for (Attribute attribute : this.attributes) {
                if (attribute.layout() == layout) {
                    return attribute;
                }
                continue;
            }
            return null;
        }
        
        public Attribute removeAttribute(String s) {
            return this.removeAttribute(this.getAttribute(s));
        }
        
        public Attribute removeAttribute(Attribute.Layout layout) {
            return this.removeAttribute(this.getAttribute(layout));
        }
        
        public void strip(String s) {
            this.removeAttribute(this.getAttribute(s));
        }
        
        static {
            noAttributes = Arrays.asList(new Attribute[0]);
        }
    }
    
    public abstract static class ValueStream
    {
        public int getInt(int n) {
            throw this.undef();
        }
        
        public void putInt(int n, int n2) {
            throw this.undef();
        }
        
        public ConstantPool.Entry getRef(int n) {
            throw this.undef();
        }
        
        public void putRef(int n, ConstantPool.Entry entry) {
            throw this.undef();
        }
        
        public int decodeBCI(int n) {
            throw this.undef();
        }
        
        public int encodeBCI(int n) {
            throw this.undef();
        }
        
        public void noteBackCall(int n) {
        }
        
        private RuntimeException undef() {
            return new UnsupportedOperationException("ValueStream method");
        }
    }
    
    public static class Layout implements Comparable<Attribute.Layout>
    {
        int ctype;
        String name;
        boolean hasRefs;
        String layout;
        int bandCount;
        Attribute.Layout.Element[] elems;
        Attribute canon;
        private static final Attribute.Layout.Element[] noElems;
        
        public int ctype() {
            return this.ctype;
        }
        
        public String name() {
            return this.name;
        }
        
        public String layout() {
            return this.layout;
        }
        
        public Attribute canonicalInstance() {
            return this.canon;
        }
        
        public ConstantPool.Entry getNameRef() {
            return ConstantPool.getUtf8Entry(this.name());
        }
        
        public boolean isEmpty() {
            return this.layout.isEmpty();
        }
        
        public Layout(int ctype, String s, String str) {
            this.ctype = ctype;
            this.name = s.intern();
            this.layout = str.intern();
            assert ctype < 4;
            int startsWith = str.startsWith("[");
            try {
                if (startsWith == 0) {
                    this.elems = Attribute.tokenizeLayout(this, -1, str);
                }
                else {
                    String[] splitBodies = Attribute.splitBodies(str);
                    Attribute.Layout.Element[] elems = new Attribute.Layout.Element[splitBodies.length];
                    this.elems = elems;
                    Object o;
                    for (int i = 0; i < elems.length; o.bandIndex = -1, o.layout = splitBodies[i], elems[i] = o, ++i) {
                        o = new Attribute.Layout.Element();
                        o.kind = 10;
                        o.removeBand();
                    }
                    for (int j = 0; j < elems.length; ++j) {
                        elems[j].body = Attribute.tokenizeLayout(this, j, splitBodies[j]);
                    }
                }
            }
            catch (StringIndexOutOfBoundsException cause) {
                throw new RuntimeException("Bad attribute layout: " + str, cause);
            }
            this.canon = new Attribute(this, Constants.noBytes);
            return;
        }
        
        private Layout() {
        }
        
        static Attribute.Layout makeKey(int ctype, String s, String s2) {
            Attribute.Layout layout = new Attribute.Layout();
            layout.ctype = ctype;
            layout.name = s.intern();
            layout.layout = s2.intern();
            assert ctype < 4;
            return layout;
        }
        
        static Attribute.Layout makeKey(int n, String s) {
            return makeKey(n, s, "");
        }
        
        public Attribute addContent(byte[] array, Object o) {
            return this.canon.addContent(array, o);
        }
        
        public Attribute addContent(byte[] array) {
            return this.canon.addContent(array, null);
        }
        
        @Override
        public boolean equals(Object o) {
            return o != null && o.getClass() == Attribute.Layout.class && this.equals((Attribute.Layout)o);
        }
        
        public boolean equals(Attribute.Layout layout) {
            return this.name.equals(layout.name) && this.layout.equals(layout.layout) && this.ctype == layout.ctype;
        }
        
        @Override
        public int hashCode() {
            return ((17 + this.name.hashCode()) * 37 + this.layout.hashCode()) * 37 + this.ctype;
        }
        
        @Override
        public int compareTo(Attribute.Layout layout) {
            int compareTo = this.name.compareTo(layout.name);
            if (compareTo != 0) {
                return compareTo;
            }
            int compareTo2 = this.layout.compareTo(layout.layout);
            if (compareTo2 != 0) {
                return compareTo2;
            }
            return this.ctype - layout.ctype;
        }
        
        @Override
        public String toString() {
            String s = Attribute.contextName(this.ctype) + "." + this.name + "[" + this.layout + "]";
            assert (s = this.stringForDebug()) != null;
            return s;
        }
        
        private String stringForDebug() {
            return Attribute.contextName(this.ctype) + "." + this.name + Arrays.asList(this.elems);
        }
        
        public boolean hasCallables() {
            return this.elems.length > 0 && this.elems[0].kind == 10;
        }
        
        public Attribute.Layout.Element[] getCallables() {
            if (this.hasCallables()) {
                return (Attribute.Layout.Element[])Arrays.copyOf(this.elems, this.elems.length);
            }
            return Attribute.Layout.noElems;
        }
        
        public Attribute.Layout.Element[] getEntryPoint() {
            if (this.hasCallables()) {
                return this.elems[0].body;
            }
            return (Attribute.Layout.Element[])Arrays.copyOf(this.elems, this.elems.length);
        }
        
        public void parse(Attribute.Holder holder, byte[] array, int n, int i, Attribute.ValueStream valueStream) {
            int using = Attribute.parseUsing(this.getEntryPoint(), holder, array, n, i, valueStream);
            if (using != n + i) {
                throw new InternalError("layout parsed " + (using - n) + " out of " + i + " bytes");
            }
            return;
        }
        
        public Object unparse(Attribute.ValueStream valueStream, ByteArrayOutputStream byteArrayOutputStream) {
            Object[] array = { null };
            Attribute.unparseUsing(this.getEntryPoint(), array, valueStream, byteArrayOutputStream);
            return array[0];
        }
        
        public String layoutForClassVersion(Package.Version version) {
            if (version.lessThan(Constants.JAVA6_MAX_CLASS_VERSION)) {
                return Attribute.expandCaseDashNotation(this.layout);
            }
            return this.layout;
        }
        
        static {
            noElems = new Attribute.Layout.Element[0];
        }
        
        public class Element
        {
            String layout;
            byte flags;
            byte kind;
            byte len;
            byte refKind;
            int bandIndex;
            int value;
            Attribute.Layout.Element[] body;
            
            boolean flagTest(byte b) {
                return (this.flags & b) != 0x0;
            }
            
            Element() {
                this.bandIndex = Attribute.Layout.this.bandCount++;
            }
            
            void removeBand() {
                Attribute.Layout this$0 = Attribute.Layout.this;
                --this$0.bandCount;
                assert this.bandIndex == Attribute.Layout.this.bandCount;
                this.bandIndex = -1;
                return;
            }
            
            public boolean hasBand() {
                return this.bandIndex >= 0;
            }
            
            @Override
            public String toString() {
                String s = this.layout;
                assert (s = this.stringForDebug()) != null;
                return s;
            }
            
            private String stringForDebug() {
                Attribute.Layout.Element[] body = this.body;
                switch (this.kind) {
                    case 9: {
                        body = null;
                        break;
                    }
                    case 8: {
                        if (this.flagTest((byte)8)) {
                            body = null;
                            break;
                        }
                        break;
                    }
                }
                return this.layout + (this.hasBand() ? ("#" + this.bandIndex) : "") + "<" + ((this.flags == 0) ? "" : ("" + (int)this.flags)) + (int)this.kind + (int)this.len + ((this.refKind == 0) ? "" : ("" + (int)this.refKind)) + ">" + ((this.value == 0) ? "" : ("(" + this.value + ")")) + ((body == null) ? "" : ("" + Arrays.asList(body)));
            }
        }
    }
    
    public static class FormatException extends IOException
    {
        private static final long serialVersionUID = -2542243830788066513L;
        private int ctype;
        private String name;
        String layout;
        
        public FormatException(String str, int ctype, String s, String layout) {
            super(Constants.ATTR_CONTEXT_NAME[ctype] + " attribute \"" + s + "\"" + ((str == null) ? "" : (": " + str)));
            this.ctype = ctype;
            this.name = s;
            this.layout = layout;
        }
        
        public FormatException(String s, int n, String s2) {
            this(s, n, s2, null);
        }
    }
}
